/*
 * Project.......: WTINOO
 * File..........: TINIO.HPP
 * Author........: Rogai Davide
 * Last review date: 24/05/2001
 * Description:
 *		Interface for the IO base-classes used by TIN objects.
 *		Interface for some useful extentions and for the containers
 */

#ifndef TINIO_HPP
#define TINIO_HPP

#include <stdio.h>
#include "thrsem.h"

#define MAX_NAMELEN 50

// IO value returned inside a IO Class

typedef enum io_val
{
  IO_ERROR=-3,
  IO_END,
  IO_NONE,
  IO_FALSE,
  IO_TRUE
} io_val;

/*
	VirtualRoundArray class: Class to manage a round array
	(is called virtual because it doesn't contain any array)
	used int Output classes and in TIN classes
*/

class VirtualRoundArray
{
private:
	unsigned int size;	//size of the time-window
	unsigned int curr;	//current entry position in the array
public:
	/* Constructor: set the size of the array */
	VirtualRoundArray(unsigned int setsize) { size=setsize; reset(); }

	/* getSize: it returns the size of the time window */
	unsigned int getSize() { return(size); }

	/* getPos: it returns the position of a given relative time (only 
	   in positive direction) getPos(0)=getCurr() */
	unsigned int getPos(unsigned int trel) { return((curr+trel)%size); }

	/* getTimeRel: it is the opposite of getPos, it returns the relative time
	   instant of the specific position in the array*/
	unsigned int getTimeRel(unsigned int pos) { return ((pos-curr+size)%size); }

	/* getCurr: it return the current entry position */
	unsigned int getCurr() { return(curr); }

	/* timeCmp: it makes a confrontation between to different position, but in
	   the sense of relative time instants */
	int timeCmp(unsigned int pos1, unsigned int pos2) 
		{ return((int)getTimeRel(pos1)-(int)getTimeRel(pos2)); }

	/* inc: it increments the time (the current entry position) */
	void inc() { curr=getPos(1); }

	/* reset: it reset the time*/
	void reset() { curr=0; }
};

/*
	NamedObj Class: Base Class for classes with a name
*/

class NamedObj
{
protected:
	char name[MAX_NAMELEN];	//name; useful to retrieve in the container
public:
	NamedObj(const char *name,const char *suffix=NULL);
	const char *getName();
};

/*
	IODev Class: Base Class for Input/Output
*/

class ChildObj
{
protected:
	void *parent;
public:
	ChildObj(void *newparent=NULL) { setParent(newparent); }
	void setParent(void *newparent) { parent=newparent; }
	void *getParent() { return(parent); }
};

class BooleanLinked
{
protected:
	bool *var;
public:
	BooleanLinked(bool *newvar=NULL) { setBooleanLink(newvar); }
	void setBooleanLink(bool *newvar) { var=newvar; }
	bool *getBooleanLink() { return(var); }
	void setLinkedValue(bool value) { *var=value; }
	bool getLinkedValue() { return(*var); }
};

class IODev: public NamedObj
{
protected:
	io_val val;	//memory variable for the result of eval
public:
	/* Constructor: it sets the name */
	IODev(const char *newname,const char *suffix=NULL);

	/* isInput, isOutput: virtual methods overrided by the imput and the output class */
	virtual bool isInput() { return(false); }
	virtual bool isOutput() { return(false); }

	/* get,set...: the following methods are created to unlink the TIN class
	   from the "io_val enum type" */
	bool getError() { return(val==IO_ERROR); }
	bool getEnd() { return(val==IO_END);  }
	bool getNone() { return(val==IO_NONE); }
	bool getTrue() { return(val==IO_TRUE); }
	bool getFalse() { return(val==IO_FALSE); }
	io_val getValue() { return(val); };

	void setError() { val=IO_ERROR; }
	void setEnd() { val=IO_END; }
	void setNone() { val=IO_NONE; }
	void setTrue() { val=IO_TRUE; }
	void setFalse() { val=IO_FALSE; }
	void setValue(io_val newval) { val=newval; };
};

/*
	Input Class: Base Class for Input
	to specialize... see "virtual input_val eval()"
*/

class Input: public IODev
{
public:
	/* Constructor: it sets the name */
	Input(const char *newname,const char *suffix=NULL);

	/* isInput: overrided for the Input class (true)*/
	bool isInput() { return(true); }

	/* eval: method overridable to create your custom Input class */
    virtual io_val eval() { return(val); }

	/* scan: used by the sincronization, it calls eval and stores the result 
	   in the val property*/
	void scan();
};

class CommandInput: public Input
{
private:
	io_val status;
public:
	/* Constructor: it provides to set the name */
	CommandInput();

	/* Constructor: it provides to set the name and to append to the name a suffix
	   useful for inputs were included in other classes */
	CommandInput(const char *newname,const char *suffix=NULL);

	/* eval(): it is not virtual now, in this particular input eval cannot do
	   anything more */
	io_val eval() { return(status); }

	/* set...: the following methods allow to set the input value asyncronously
	   by calling in command line (to be included in other classes */
	void setError() { status=IO_ERROR; }
	void setEnd() { status=IO_END; }
	void setNone() { status=IO_NONE; }
	void setFalse() { status=IO_FALSE; }
	void setTrue() { status=IO_TRUE; }
};

/*
	Output Class: Base Class for Output
	to specialize... see "virtual void action(bool value)"
*/

class Output: public IODev
{
protected:
	bool closed;	//closed flag (when it is closed no more change will be done)

public:
	/* Contructor: set the name and the dimension of the time window */
	Output(const char *newname, const char  *suffix=NULL);

	/* isOutput: overrided (see Input::isInput) */
	virtual bool isOutput() { return(true); }

	/*  isClosed: it returns if the Output is closed */
	bool isClosed() { return(closed); }

	/* set to IO_NONE every position in the time-window */
	void reset();

	/* change: usually it performs action with the current value, but could be
	   ovverided to  perform more complex action (see the inherited classes) */
	virtual void change();
	/* action: overridable method to perform your custom Output */
	virtual void action(bool value) {}
	/* onError: overridable method to perform your custom Output */
	virtual void onError() {}
	/* onNone: overridable method to perform your custom Output */
	virtual void onNone() {}
	/* onTrue: overridable method to perform your custom Output */
	virtual void onTrue() {}
	/* onFalse: overridable method to perform your custom Output */
	virtual void onFalse() {}

	/* close: as change in this base class it sets only the closed flag, but 
	   could be inherited to send close signal (see the inherited classes) */
	virtual void close() { closed=true; }
};

/*
	FeedbackOut Class: Base Class for Output Feedback (it contains an input linkable
	as an Input device to specialize... see "virtual void action(bool value)"
*/

class FeedbackOutput: public Output
{
private:

	CommandInput feedback;	//Feedback input device

public:
	/* Constructor: nothing different from Output, it only sets the right name of the
	   internal Input Device */
	FeedbackOutput(const char *newname);

	/* getFeedback: return the Input device, is used to link the device in the container*/
	Input *getFeedback() { return(&feedback); }
		
	/* change: this is important for this kind of output, here, after a call to the 
	   superclass method it rewrites the value on the Feedback input that will be scanned
	   at the next time instant */
	void change();

	/* close: as change it adds to the normal close procedure the signal of close to the
	   related feedback input */
	virtual void close();
};

/*
	ThreadOutput Class: Base Class for Output that run on a separate thread
	to specialize... see "virtual void action(bool value)" but the value will 
	be always set to true.
*/

class ThreadOutput : public Output , protected Thread
{
protected:
	bool onrun;		//onrun flag: set to false by the thread when it's finish

private:
	/* run: here is defined the instruction for the thread, it calls the action with
	   the true value, it makes a loop to enable several launches of the thread */
	int run();

public:
	/* Constructor: it creates the thread */
	ThreadOutput(const char *newname);
	
	/* close: not yet implemented, it should terminate the thread */
	//virtual void close() {}
	
	/* change: in this case the change method controls the thread start*/
	void change();
};

/*
	ThreadOutput Class: Base Class for Output that run on a separate thread
	that con be controlled in time and put the result on two different input
	devices, one will be true if the thread was terminated before the time-out
	the other if the thread doesn't fit in the dedicated time
*/

class BoundedThreadOutput : public ThreadOutput
{
private:
	unsigned int time_out;	//Time-out expressed in delta
	unsigned int run_count;	//Counter for the running mode
	bool started;	//started flag: set to on when the thread starts

	CommandInput ok;	//ok is set to true if the thread finish before the time-out
	CommandInput fail;	//fail is set to true if the thread run on time-out

public:
	/* Constructor: it sets also the time-out */
	BoundedThreadOutput(const char *newname,unsigned int newtimeout);

	/* getOk, getFail: necessary to link these internal input into containers or with TIN */
	Input *getOk() { return(&ok); }
	Input *getFail() { return(&fail); }

	/* close: in this case it sends the close signal to the Ok and Fail input */
	virtual void close() ;

	/* change: in this case is a little more complex, because is the responsable of the
	   time control of  the thread */
	void change();
};

class BooleanLinkedInput : public Input, public BooleanLinked
{
public:
	BooleanLinkedInput(const char *newname, const char *suffix=NULL, bool *newvar=NULL)
		: Input(newname,suffix), BooleanLinked(newvar) {}
	io_val eval() { return(getLinkedValue()?IO_TRUE:IO_FALSE); }
};

class BooleanLinkedOutput : public Output, public BooleanLinked
{
public:
	BooleanLinkedOutput(const char *newname, const char *suffix=NULL, bool *newvar=NULL)
		: Output(newname,suffix), BooleanLinked(newvar) {}
	void action(bool value) { setLinkedValue(value); }
};

class ThresholdOnDoubleInput : public Input
{
protected:
	double *var;
	double threshold;
public:
	ThresholdOnDoubleInput(const char *newname, const char *suffix=NULL, 
			double *newvar=NULL, double newthreshold=0.0) : Input(newname, suffix) 
			{ setDoubleLink(newvar); setThreshold(newthreshold); }
	void setDoubleLink(double *newvar) { var=newvar; }
	double *getDoubleLink() { return(var); }
	double getLinkedValue() { return(*var); }
	void setThreshold(double newthreshold) { threshold=newthreshold; }
	double getThreshold() { return(threshold); }
	io_val eval() { return((getLinkedValue()>threshold)?IO_TRUE:IO_FALSE); }
};

/*	
	*************
	IOBox Classes: IODev Containers to synchronize the IO operations
	*************
*/

/*
	Box Class: base class for the containers (array of void *)
*/

class Box
{
protected:
	int maxitems;			// Max items linkable to the Box
	int linked;				// Objects currently linked
	NamedObj **link_arr;	// general classes pointers dynamic array

public:
	/* Constructor: it fixes the size of the Box */
	Box(int max);
	~Box() { delete(link_arr); }

	/* add: add a pointer to the Box */
	void add(NamedObj *newobj);

	/* getItem: it searchs an object called "obj_name" and returns the pointer */
	NamedObj *getItem(const char *obj_name);
};

class InBox : public Box
{
private:
	Mutex scanlock;
public:
	/* Constructor: it fixes the size of the InBox */
	InBox(int max);

	/* add: it contains only Input class pointers */
	void add(Input *newinput);

	/* getItem: it searchs the Input called "in_name" and returns the pointer */
	Input *getItem(const char *in_name);

	/* scan: it scans all the inputs of the InBox */
	void scan();

	void beginScan() { /*scanlock.take();*/ }
	void endScan() { /*scanlock.give();*/ }
};

class OutBox : public Box
{
public:
	/* Constructor: it fixes the size of the OuBox */
	OutBox(int max);

	/* add: contains only Input class pointers */
	void add(Output *newoutput);
	
	/* getItem: it searchs the Output called "out_name" and returns the pointer */
	Output *getItem(const char *out_name);

	/* change: it changes all the outputs of the OutBox */
	void change();
	
	/* isClosed: it returns if all the output linked are closed */
	bool isClosed();
};

#endif //TINIO_HPP
