/*
 * Project.......: WTINOO
 * File..........: TINOO.HPP
 * Author........: Rogai Davide
 * Last review date: 24/05/2001
 * Description:
 *		Interface for the TIN classes. This is an embedding of the
 *		Temporal Inference Network executor designed by Giotti (C-code).
 */


#ifndef TINOO_HPP
#define TINOO_HPP 

#include "tinx.h"
#include "thrsem.h"
#include <stdio.h>

/*
	TIN Class: embedding of the Temporal Inference Network it runs on a separate thread
*/

class TIN : public NamedObj, private Thread
{

	/* kb: Knowledge base: Struct built to perform the Inference...
	   it chooses the events to process and it processes these */
	k_base *kb;
	
	bool alive;			//flag to determine if there is some new event to process
	bool oninference;

	/* run: in these cae the thread method is overrided to perform the inference loop 
	   until the time instant has no more events to process */
	int run();

public:
	/* readInputs: it has to  operate after the scan call to the inputs, it reads the 
	   stored value and generate the new events to process, it returns if there is
	   some inputs that are still open */
	void readInputs();
	void writeOutputs();

	/* closeOutputs: this method is called when the TIN end is life-cycle, it provides
	   to close all the outputs connected to it */
	void closeOutputs();

	/* errorOutputs: this method is called when an input occurs on the TIN, it provides
	   to send an Error Message to all the outputs connected */
	void errorOutputs();
    
	/* Constructor: it sets the name, it open the knowledge_base by passing the filename
	   (withou extention) and the time-horizon (the most far future instant processable) */
	TIN(const char *newname, const char *basename, int newtw_size);
	/* Constructor: it sets the name, it open the knowledge_base by passing the same name
	   for the TIN file and the initial conditions */
	TIN(const char *newname, int newtw_size);

	/* Destructor: it provides to free the memory allocation */
	~TIN();

	/* loadKnowBase, loadInitState: These method read from a file the network
	   and the initial condition of it, they work with special file (.TIN, .EVL) */
	void loadKnowBase(const char *basename, int neww_size);
	void loadInitState(char *statename);
	void setInferenceHorizon(unsigned int horiz);

	/* inc/resetTime: they control the time-window position, the time window is 
	   allocated in the knowledge base data structure */
	void resetTime();
	void setInputStartTime(unsigned int start=0);

	/* createThread: it creates the thread ready to start */
	void createThread();

	/* runInference: this method is called by the syncronization, it starts the thread 
	   and let the network to process the new events aquired by the inputs */
	void wakeUpInference();

	void makeAllPossibleInference();

	/* checkTerminated: called by the syncronizer, it check if all the events were
	   processed (usually within delta miliseconds) */
	bool checkTerminated();

	/* isAlive: returns if the TIN has something more to process  */
	bool isAlive();
	
	/* linkIODev: this method link the old data structure with the new Object-Oriented
	   conception of the TIN (IO Devices) */
	void linkIODev(IODev *io, const char *ioname);

	/* listIO, checkIO: debug method, to list the IOs of a loaded TIN, and to check the
	   connections just made */
	int countInputs();
	int countOutputs();
	void listIO();
	void listIO(char **in_names, char **out_names);

    bool checkIO();

	void dumpIOs(FILE *filelog=NULL);
};

/*
	TINBox class: a TIN container to allow synchronized operation with mamy TIN objects.
*/

class TINBox : public Box
{
public:
	/* Constructor: it sets the max items in the container */
	TINBox(int max);

	/* add: it adds a new TIN in the container */
	void add(TIN *newtin);

	/* getItem: it searchs the TIN called "tin_name" and returns the pointer */
	TIN *getItem(const char *tin_name);

	/* inc/resetTime: they perform time operation for all the items contained */
	void resetTime();

	/* createThread, runInference: they provide to call the same method to every item */
	void createThread();

	void wakeUpInference();
	void makeAllPossibleInference();

	void readInputs();
	void writeOutputs();

	void dumpIOs(FILE *filelog=NULL);

	/* checkTerminated, isAlive: these return a collective property, if every TIN
	   has terminated, and if at least one TIN is still alive */
	bool checkTerminated();
	bool isAlive();
};

#endif //TINOO_HPP
